diff --git a/src/backend/access/transam/xact.c b/src/backend/access/transam/xact.c
index 9328d26..565c41b 100644
--- a/src/backend/access/transam/xact.c
+++ b/src/backend/access/transam/xact.c
@@ -914,7 +914,6 @@ ExitParallelMode(void)
 	TransactionState s = CurrentTransactionState;
 
 	Assert(s->parallelModeLevel > 0);
-	Assert(s->parallelModeLevel > 1 || !ParallelContextActive());
 
 	--s->parallelModeLevel;
 }
@@ -935,6 +934,15 @@ IsInParallelMode(void)
 }
 
 /*
+ *	getParallelModeLevel
+ */
+bool
+getParallelModeLevel(void)
+{
+	return CurrentTransactionState->parallelModeLevel;
+}
+
+/*
  *	CommandCounterIncrement
  */
 void
diff --git a/src/backend/executor/execMain.c b/src/backend/executor/execMain.c
index 7d57f8c..762a526 100644
--- a/src/backend/executor/execMain.c
+++ b/src/backend/executor/execMain.c
@@ -1461,6 +1461,8 @@ ExecEndPlan(PlanState *planstate, EState *estate)
 		ExecEndNode(subplanstate);
 	}
 
+	if (estate->es_use_parallel_mode)
+		Assert (getParallelModeLevel() > 0 || !ParallelContextActive());
 	/*
 	 * destroy the executor's tuple table.  Actually we only care about
 	 * releasing buffer pins and tupdesc refcounts; there's no need to pfree
diff --git a/src/backend/executor/execParallel.c b/src/backend/executor/execParallel.c
index 4772408..522b481 100644
--- a/src/backend/executor/execParallel.c
+++ b/src/backend/executor/execParallel.c
@@ -569,13 +569,6 @@ ExecParallelCleanup(ParallelExecutorInfo *pei)
 	if (pei->instrumentation)
 		ExecParallelRetrieveInstrumentation(pei->planstate,
 											pei->instrumentation);
-
-	if (pei->pcxt != NULL)
-	{
-		DestroyParallelContext(pei->pcxt);
-		pei->pcxt = NULL;
-	}
-	pfree(pei);
 }
 
 /*
diff --git a/src/backend/executor/nodeGather.c b/src/backend/executor/nodeGather.c
index 216759a..5fc9233 100644
--- a/src/backend/executor/nodeGather.c
+++ b/src/backend/executor/nodeGather.c
@@ -261,7 +261,21 @@ ExecGather(GatherState *node)
 void
 ExecEndGather(GatherState *node)
 {
+	ParallelExecutorInfo *pei = node->pei;
 	ExecShutdownGather(node);
+
+	if (pei != NULL)
+	{
+		if (pei->pcxt != NULL)
+		{
+			DestroyParallelContext(pei->pcxt);
+			pei->pcxt = NULL;
+		}
+
+		pfree(pei);
+		node->pei = NULL;
+	}
+
 	ExecFreeExprContext(&node->ps);
 	ExecClearTuple(node->ps.ps_ResultTupleSlot);
 	ExecEndNode(outerPlanState(node));
@@ -441,7 +455,6 @@ ExecShutdownGather(GatherState *node)
 	if (node->pei != NULL)
 	{
 		ExecParallelCleanup(node->pei);
-		node->pei = NULL;
 	}
 }
 
diff --git a/src/include/access/xact.h b/src/include/access/xact.h
index 243811c..83d90f3 100644
--- a/src/include/access/xact.h
+++ b/src/include/access/xact.h
@@ -384,5 +384,6 @@ extern void ParseAbortRecord(uint8 info, xl_xact_abort *xlrec, xl_xact_parsed_ab
 extern void EnterParallelMode(void);
 extern void ExitParallelMode(void);
 extern bool IsInParallelMode(void);
+extern bool getParallelModeLevel(void);
 
 #endif   /* XACT_H */
